#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from django.contrib.gis.geos import GEOSGeometry, Point, Polygon, MultiPolygon
from django.contrib.gis.measure import D

from eoxserver.core.decoders import kvp, enum
from eoxserver.core.util.xmltools import NameSpace


class GeoExtension(object):
    """ Implementation of the OpenSearch `'Geo' extension draft
    <http://www.opensearch.org/Specifications/OpenSearch/Extensions/Geo/1.0/Draft_2>`_.
    Currently all parameters apart from the ``name`` are supported. The point
    plus radius with the relation type ``contains`` requires a PostGIS database
    backend.
    """

    namespace = NameSpace(
        "http://a9.com/-/opensearch/extensions/geo/1.0/", "geo"
    )

    def filter(self, qs, parameters):
        decoder = GeoExtensionDecoder(parameters)

        geom = decoder.box or decoder.geometry
        lon, lat = decoder.lon, decoder.lat
        distance = decoder.radius
        relation = decoder.relation
        uid = decoder.uid

        if geom:
            if relation == "intersects":
                qs = qs.filter(footprint__intersects=geom)
            elif relation == "contains":
                qs = qs.filter(footprint__coveredby=geom)
            elif relation == "disjoint":
                qs = qs.filter(footprint__disjoint=geom)

        elif lon is not None and lat is not None and distance is not None:
            geom = Point(lon, lat)
            if relation == "intersects":
                qs = qs.filter(footprint__distance_lte=(geom, distance))
            elif relation == "contains":
                # TODO: right?, also only available on postgis
                qs = qs.filter(footprint__dwithin=(geom, distance))
            elif relation == "disjoint":
                qs = qs.filter(footprint__distance_gt=(geom, distance))
        elif lon is not None and lat is not None:
            geom = Point(lon, lat)
            if relation == "intersects":
                qs = qs.filter(footprint__intersects=geom)
            elif relation == "contains":
                qs = qs.filter(footprint__coveredby=geom)
            elif relation == "disjoint":
                qs = qs.filter(footprint__disjoint=geom)

        if uid:
            qs = qs.filter(identifier=uid)

        return qs

    def get_schema(self, collection=None, model_class=None):
        return (
            dict(name="bbox", type="box"),
            dict(name="geom", type="geometry", profiles=[
                dict(
                    href="http://www.opengis.net/wkt/LINESTRING",
                    title="This service accepts WKT LineStrings"
                ),
                dict(
                    href="http://www.opengis.net/wkt/POINT",
                    title="This service accepts WKT Point"
                ),
                dict(
                    href="http://www.opengis.net/wkt/POLYGON",
                    title="This service accepts WKT Polygons"
                ),
                dict(
                    href="http://www.opengis.net/wkt/MULTILINESTRING",
                    title="This service accepts WKT Multi-LineStrings"
                ),
                dict(
                    href="http://www.opengis.net/wkt/MULTIPOINT",
                    title="This service accepts WKT Multi-Point"
                ),
                dict(
                    href="http://www.opengis.net/wkt/MULTIPOLYGON",
                    title="This service accepts WKT Multi-Polygons"
                ),
            ]),
            dict(name="lon", type="lon"),
            dict(name="lat", type="lat"),
            dict(name="r", type="radius"),
            dict(name="georel", type="relation",
                options=["intersects", "contains", "disjoint"]
            ),
            dict(name="uid", type="uid")
        )


def parse_bbox(raw):
    values = list(map(float, raw.split(",")))
    if len(values) != 4:
        raise ValueError("Invalid number of coordinates in 'bbox'.")

    minx, miny, maxx, maxy = values
    if minx <= maxx:
        return Polygon.from_bbox(values)

    return MultiPolygon(
        Polygon.from_bbox((minx, miny, 180.0, maxy)),
        Polygon.from_bbox((-180.0, miny, maxx, maxy)),
    )


def parse_radius(raw):
    # TODO: allow the specification of additional units
    value = float(raw)
    if value < 0:
        raise ValueError("Invalid radius specified")
    return D(m=value)


class GeoExtensionDecoder(kvp.Decoder):
    box = kvp.Parameter(num="?", type=parse_bbox)
    radius = kvp.Parameter(num="?", type=float)
    geometry = kvp.Parameter(num="?", type=GEOSGeometry)
    lon = kvp.Parameter(num="?", type=float)
    lat = kvp.Parameter(num="?", type=float)
    relation = kvp.Parameter(num="?",
        type=enum(("intersects", "contains", "disjoint"), False),
        default="intersects"
    )
    uid = kvp.Parameter(num="?")



#-------------------------------------------------------------------------------
#
#  WPS Complex Data type
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
# pylint: disable=too-few-public-methods,

import json
from os import remove
from copy import deepcopy

try:
    from StringIO import StringIO
    from cStringIO import StringIO as FastStringIO
    BytesIO = StringIO
except ImportError:
    from io import BytesIO
    from io import StringIO
    from io import StringIO as FastStringIO

try:
    # available in Python 2.7+
    from collections import OrderedDict
except ImportError:
    from django.utils.datastructures import SortedDict as OrderedDict

from lxml import etree
from .base import Parameter
from .formats import Format
from django.utils.encoding import smart_text
from django.utils.six import string_types, text_type, itervalues, binary_type

#-------------------------------------------------------------------------------
# complex data - data containers

class CDBase(object):
    """ Base class of the complex data container.

    Constructor parameters (all optional and all defaulting to None):
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        headers    additional raw output HTTP headers encoded as a list
                   of <key>, <value> pairs (tuples).
    """
    def __init__(self, mime_type=None, encoding=None, schema=None, format=None,
                 filename=None, headers=None, **kwargs):
        # pylint: disable=redefined-builtin, unused-argument, too-many-arguments
        if isinstance(format, Format):
            self.mime_type = format.mime_type
            self._encoding = format.encoding
            self.schema = format.schema
        else:
            self.mime_type = mime_type
            self._encoding = encoding
            self.schema = schema
        self.filename = filename
        self.headers = headers or []

    @property
    def encoding(self):
        return self._encoding

    @property
    def data(self):
        """ Get the payload data. """
        raise NotImplementedError


class CDObject(CDBase):
    """ Complex data wrapper around an arbitrary python object.
        To be used to set custom format attributes for the XML and JSON payload.
        NOTE: CDObject is not used for the input JSON and XML.

    Constructor parameters:
        data       mandatory object holding the payload data
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        headers    additional raw output HTTP headers encoded as a list
                   of <key>, <value> pairs (tuples).
    """
    def __init__(self, data, *args, **kwargs):
        CDBase.__init__(self, *args, **kwargs)
        self._data = data

    @property
    def data(self):
        return self._data


class CDByteBuffer(BytesIO, CDBase):
    """ Complex data binary in-memory buffer (StringIO).
        To be used to hold a generic binary (byte-stream) payload.

    Constructor parameters:
        data       optional initial payload byte string
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        headers    additional raw output HTTP headers encoded as a list
                   of <key>, <value> pairs (tuples).
    """
    def __init__(self, data=b'', *args, **kwargs):
        # NOTE: StringIO is an old-style class and super cannot be used!
        BytesIO.__init__(self, data)
        CDBase.__init__(self, *args, **kwargs)

    def write(self, data):
        BytesIO.write(self, data)

    @property
    def data(self):
        self.seek(0)
        return self.read()


class CDTextBuffer(StringIO, CDBase):
    """ Complex data text (Unicode) in-memory buffer (StringIO).
        To be used to hold generic text. The text payload is stored as
        a Unicode-stream.

    Constructor parameters:
        data       optional initial payload Unicode string
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        headers    additional raw output HTTP headers encoded as a list
                   of <key>, <value> pairs (tuples).
        text_encoding   optional keyword parameter defining the input text
                   encoding. By default UTF-8 is assumed.
    """
    def __init__(self, data=u'', *args, **kwargs):
        # NOTE: StringIO is an old-style class and super cannot be used!
        StringIO.__init__(self, smart_text(data))
        CDBase.__init__(self, *args, **kwargs)
        self.text_encoding = kwargs.get('text_encoding', None)

    @property
    def data(self):
        self.seek(0)
        return self.read()

    def write(self, data):
        if self.text_encoding is None:
            return StringIO.write(self, smart_text(data))
        else:
            return StringIO.write(self, smart_text(data, self.text_encoding))

    def read(self, size=None):
        if size is None:
            data = StringIO.read(self)
        else:
            data = StringIO.read(self, size)
        if self.text_encoding is not None:
            data = data.encode(self.text_encoding)
        return data


class CDAsciiTextBuffer(CDByteBuffer):
    """ Complex data text (ASCII) in-memory buffer (StringIO).
        To be used to hold generic ASCII text. The text payload
        is stored as a byte-stream and this class cannot hold
        characters outside of the 7-bit ASCII characters' range.

    Constructor parameters:
        data       optional initial payload ASCII string
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        headers    additional raw output HTTP headers encoded as a list
                   of <key>, <value> pairs (tuples).
        text_encoding   optional keyword parameter defining the input text
                   encoding. By default ASCII is assumed.
    """
    def __init__(self, data='', *args, **kwargs):
        CDByteBuffer.__init__(self, data.encode('ascii'), *args, **kwargs)
        self.text_encoding = kwargs.get('text_encoding', None)

    def write(self, data):
        if not isinstance(data, string_types):
            data = str(data)
        CDByteBuffer.write(self, data.encode('ascii'))

    def read(self, size=None):
        if size is None:
            data = CDByteBuffer.read(self)
        else:
            data = CDByteBuffer.read(self, size)
        if self.text_encoding not in ('ascii', 'utf-8'): # ASCII is a subset of UTF-8
            data = data.decode('ascii')
            if self.text_encoding is not None:
                data = data.encode(self.text_encoding)
        return data


class CDFileWrapper(CDBase):
    """ Complex data file (or file-like) object wrapper.

    Constructor parameters:
        file_object  mandatory seekable Python file or file-like object.
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        text_encoding optional source text file encoding
    """

    def __init__(self, file_object, *args, **kwargs):
        CDBase.__init__(self, *args, **kwargs)
        self._file = file_object
        self.text_encoding = kwargs.get('text_encoding', None)

    def __del__(self):
        if hasattr(self, "_file"):
            self.close()

    @property
    def data(self):
        self.seek(0)
        return self.read()

    def __getattr__(self, attr):
        if attr == "_file":
            raise AttributeError("Instance has no attribute '_file'")
        else:
            # Allow object to behave like a file.
            return getattr(self._file, attr)


class CDFile(CDFileWrapper):
    """ Complex data file.
        To be used to hold a generic (binary or text) byte-stream payload.
        NOTE: The file allows you to specify whether the file is
              temporary (will be automatically removed - by default)
              or permanent (preserved after object destruction).

    Constructor parameters:
        name       mandatory file-name
        mode       opening mode (passed to `open`, 'r' by default)
        buffering  buffering mode (passed to `open`, -1 by default)
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
        remove_file  optional  keyword argument defining whether the file
                   should be removed or not. Set to True by default.
    """

    def __init__(self, name, mode='rb', buffering=-1, *args, **kwargs):
        CDFileWrapper.__init__(
            self, open(name, mode, buffering), *args, **kwargs
        )
        self._remove_file = kwargs.get('remove_file', True)

    def __del__(self):
        if hasattr(self, "_file"):
            name = self.name
            self.close()
            if self._remove_file:
                remove(name)


class CDPermanentFile(CDFile):
    """ Complex data permanent file.
        To be used to hold a generic (binary or text) byte-stream payload.
        NOTE: This class preserves the actual file.

    Constructor parameters:
        name       mandatory file-name
        mode       opening mode (passed to `open`, 'r' by default)
        buffering  buffering mode (passed to `open`, -1 by default)
        mime_type  ComplexData mime-type
        encoding   ComplexData encoding
        schema     ComplexData XML schema (applicable XML only)
        format     an alternative format object defining the ComplexData
                   mime_type, encoding, and XML schema
        filename   optional raw output file-name set in the Content-Disposition
                   HTTP header.
    """

    def __init__(self, *args, **kwargs):
        kwargs['remove_file'] = False
        CDFile.__init__(self, *args, **kwargs)

#-------------------------------------------------------------------------------

class ComplexData(Parameter):
    """ Complex-data parameter class

    Constructor parameters:
        identifier  identifier of the parameter.
        title       optional human-readable name (defaults to identifier).
        abstract    optional human-readable verbose description.
        metadata    optional metadata (title/URL dictionary).
        optional    optional boolean flag indicating whether the input
                    parameter is optional or not.
        formats     List of supported formats.
        resolve_input_references Set this option to False not to resolve
                    input references. By default the references are
                    resolved (downloaded and parsed) transparently.
                    If set to False the references must be handled
                    by the process.
    """

    def __init__(self, identifier, formats, *args, **kwargs):
        super(ComplexData, self).__init__(identifier, *args, **kwargs)
        self.formats = OrderedDict()
        if isinstance(formats, Format):
            formats = (formats,)
        for frm in formats:
            self.formats[(frm.mime_type, frm.encoding, frm.schema)] = frm

    @property
    def default_format(self):
        """ Get default the default format. """
        return next(itervalues(self.formats))

    def get_format(self, mime_type, encoding=None, schema=None):
        """ Get format definition for the given mime-type and the optional
        encoding and schema.
        """
        if mime_type is None:
            return self.default_format
        else:
            return self.formats.get((mime_type, encoding, schema))

#    def verify_format(self, format):
#        """ Returns valid format or rise value error exception."""
#        if format is None:
#            return self.default_format
#        tmp = (format.mime_type, format.encoding, format.schema)
#        if tmp in self.formats:
#            return format
#        raise ValueError("Invalid format %r" % format)

    def parse(self, data, mime_type, schema, encoding, **opt):
        """ parse input complex data """
        format_ = self.get_format(mime_type, encoding, schema)
        if format_ is None:
            raise ValueError(
                "Invalid format specification! mime_type=%r, "
                "encoding=%r, schema=%r" % (mime_type, encoding, schema)
            )
        text_encoding = getattr(format_, 'text_encoding', 'utf-8')
        fattr = {
            'mime_type': format_.mime_type,
            'encoding': format_.encoding,
            'schema': format_.schema
        }
        if format_.is_xml:
            parsed_data = CDObject(etree.fromstring(data), **fattr)
        elif format_.is_json:
            parsed_data = CDObject(
                json.loads(_unicode(data, text_encoding)), **fattr
            )
        elif format_.is_text:
            # pylint: disable=redefined-variable-type
            parsed_data = CDTextBuffer(_unicode(data, text_encoding), **fattr)
            parsed_data.seek(0)
        else: # generic binary byte-stream
            parsed_data = CDByteBuffer(data, **fattr)
            if format_.encoding is not None:
                data_out = BytesIO()
                for chunk in format_.decode(parsed_data, **opt):
                    if isinstance(chunk, binary_type):
                        chunk = chunk.decode('utf-8')
                    data_out.write(chunk)
                parsed_data = data_out
            parsed_data.seek(0)
        return parsed_data

    def encode_xml(self, data):
        """ encode complex data to be embedded to an XML document"""
        mime_type = getattr(data, 'mime_type', None)
        encoding = getattr(data, 'encoding', None)
        schema = getattr(data, 'schema', None)
        format_ = self.get_format(mime_type, encoding, schema)
        if format_ is None:
            raise ValueError(
                "Invalid format specification! mime_type=%r, "
                "encoding=%r, schema=%r" % (mime_type, encoding, schema)
            )
        if not format_.allows_xml_embedding:
            raise ValueError(
                "Selected format does not allows XML embedding! mime_type=%r, "
                "encoding=%r, schema=%r" % (mime_type, encoding, schema)
            )
        if isinstance(data, CDObject):
            data = data.data
        if format_.is_xml:
            if isinstance(data, etree._ElementTree):
                data = data.getroot()
            return deepcopy(data)
        elif format_.is_json:
            return json.dumps(data, ensure_ascii=False)
        elif format_.is_text:
            if not isinstance(data, string_types):
                data.seek(0)
                data = data.read()
            return data
        else: # generic binary byte-stream
            if format_.encoding is not None:
                data.seek(0)
                data_out = FastStringIO()
                # data_out.write(str(data.data,'utf-8'))
                for chunk in format_.encode(data):
                    if isinstance(chunk, binary_type):
                        chunk = chunk.decode('utf-8')

                    data_out.write(chunk)
                data = data_out
            data.seek(0)
            return data.read()

    def encode_raw(self, data):
        """ encode complex data for raw output """
        def _rewind(fid):
            if hasattr(fid, 'seek'):
                fid.seek(0)
            return data
        mime_type = getattr(data, 'mime_type', None)
        encoding = getattr(data, 'encoding', None)
        schema = getattr(data, 'schema', None)
        format_ = self.get_format(mime_type, encoding, schema)
        text_encoding = getattr(format_, 'text_encoding', 'utf-8')
        if format_ is None:
            raise ValueError(
                "Invalid format specification! mime_type=%r, "
                "encoding=%r, schema=%r" % (mime_type, encoding, schema)
            )
        if isinstance(data, CDObject):
            data = data.data
        if format_.is_xml:
            data = BytesIO(etree.tostring(
                data, pretty_print=False, xml_declaration=True,
                encoding=text_encoding
            ))
            content_type = "%s; charset=%s" % (
                format_.mime_type, text_encoding
            )
        elif format_.is_json:
            data = BytesIO(
                json.dumps(data, ensure_ascii=False).encode(text_encoding)
            )
            content_type = "%s; charset=%s" % (
                format_.mime_type, text_encoding
            )
        elif format_.is_text:
            if isinstance(data, (CDTextBuffer, CDAsciiTextBuffer)):
                data.text_encoding = text_encoding
            else:
                source_text_encoding = getattr(data, 'text_encoding', None)
                if source_text_encoding != text_encoding:
                    data = _rewind(data).read()
                    if source_text_encoding is not None:
                        data = data.decode(source_text_encoding)
                    data = BytesIO(data.encode(text_encoding))
            content_type = "%s; charset=%s" % (
                format_.mime_type, text_encoding
            )
        else:  # generic binary byte-stream
            if format_.encoding is not None:
                data_out = BytesIO()
                for chunk in format_.encode(_rewind(data)):
                    # if isinstance(chunk, binary_type):
                    #     chunk = str(chunk,'utf-8')
                    data_out.write(chunk)
                data = data_out
            content_type = format_.mime_type
        return _rewind(data), content_type


def _bytestring(data):
    if isinstance(data, str):
        return data
    raise TypeError("Byte string expected, %s received!" % type(data))


def _unicode(data, encoding):

    if isinstance(data, text_type):
        return data
    elif isinstance(data, bytes):
        return smart_text(data, encoding)
    raise TypeError(
        "Byte or Unicode string expected, %s received!" % type(data)
    )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

""" This module contains facilities to help decoding XML structures.
"""

from lxml import etree
from eoxserver.core.decoders.base import BaseParameter
from django.utils.six import string_types


class Parameter(BaseParameter):
    """ Parameter for XML values.

        :param selector: the node selector; if a string is passed it is
                         interpreted as an XPath expression, a callable will be
                         called with the root of the element tree and shall
                         yield any number of node
        :param type: the type to parse the raw value; by default the raw
                     string is returned
        :param num: defines how many times the key can be present; use any
                    numeric value to set it to a fixed count, "*" for any
                    number, "?" for zero or one time or "+" for one or more
                    times
        :param default: the default value
        :param namespaces: any namespace necessary for the XPath expression;
                           defaults to the :class:`Decoder` namespaces.
        :param locator: override the locator in case of exceptions
    """

    def __init__(self, selector, type=None, num=1, default=None,
                 namespaces=None, locator=None):
        super(Parameter, self).__init__(type, num, default)
        self.selector = selector
        self.namespaces = namespaces
        self._locator = locator

    def select(self, decoder):
        # prepare the XPath selector if necessary
        if isinstance(self.selector, string_types):
            namespaces = self.namespaces or decoder.namespaces
            self.selector = etree.XPath(self.selector, namespaces=namespaces)
        results = self.selector(decoder._tree)
        if isinstance(results, (string_types + (float, int))):
            results = [results]

        return results

    @property
    def locator(self):
        return self._locator or str(self.selector)


class Decoder(object):
    """ Base class for XML Decoders.

        :param params: an instance of either :class:`lxml.etree.ElementTree`,
                       or :class:`basestring` (which will be parsed using
                       :func:`lxml.etree.fromstring`)

    Decoders should be used as such:
    ::

        from eoxserver.core.decoders import xml
        from eoxserver.core.decoders import typelist

        class ExampleDecoder(xml.Decoder):
            namespaces = {"myns": "http://myns.org"}
            single = xml.Parameter("myns:single/text()", num=1)
            items = xml.Parameter("myns:collection/myns:item/text()", num="+")
            attr_a = xml.Parameter("myns:object/@attrA", num="?")
            attr_b = xml.Parameter("myns:object/@attrB", num="?", default="x")


        decoder = ExampleDecoder('''
            <myns:root xmlns:myns="http://myns.org">
                <myns:single>value</myns:single>
                <myns:collection>
                    <myns:item>a</myns:item>
                    <myns:item>b</myns:item>
                    <myns:item>c</myns:item>
                </myns:collection>
                <myns:object attrA="value"/>
            </myns:root>
        ''')

        print decoder.single
        print decoder.items
        print decoder.attr_a
        print decoder.attr_b
    """

    namespaces = {}  # must be overriden if the XPath expressions use namespaces

    def __init__(self, tree):
        if isinstance(tree, string_types) or isinstance(tree, bytes):
            try:
                tree = etree.fromstring(tree)
                
            except etree.XMLSyntaxError as exc:
                # NOTE: lxml.etree.XMLSyntaxError is incorretly identified as
                #       an OWS exception by the exception handler leading
                #       to a wrong OWS error response.  This exception thus
                #       must be cought and replaced by another exception
                #       of a different type.
                raise ValueError("Malformed XML document. Error was %s" % exc)
        self._tree = tree



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""\
The eoxserver.core package provides functionality for the initialization and
re-initialization of the component system.
For convenience, the module imports the most important items from the
:mod:`eoxserver.core.component` module and instantiates a component manager
:obj:`eoxserver.core.env`.
"""


import logging
import threading

from eoxserver.core.component import (
    ComponentManager, ComponentMeta, Component,
    ExtensionPoint, UniqueExtensionPoint, implements
)
from eoxserver.core.util.importtools import easy_import, import_module


env = ComponentManager()
logger = logging.getLogger(__name__)

__init_lock = threading.RLock()
__is_initialized = False


def initialize():
    """ Initialize the EOxServer plugin system by trying to import all the
        plugins referenced by the `PLUGINS` configuration entry from the
        settings module. If a module path ends with '*' then all direct
        submodules will be imported aswell and if it ends with '**' it means
        that the import will be done recursively.
    """

    global __is_initialized

    with __init_lock:
        if __is_initialized:
            return
        __is_initialized = True

        from django.conf import settings

        logger.info("Initializing EOxServer components.")

        for plugin in getattr(settings, "COMPONENTS", ()):
            easy_import(plugin)


def reset():
    """ Reset the EOxServer plugin system.
    """

    global __is_initialized

    with __init_lock:
        if not __is_initialized:
            return
        __is_initialized = False

        logger.info("Resetting EOxServer components.")
        ComponentMeta._registry = {}
        ComponentMeta._components = []

        initialize()



#-------------------------------------------------------------------------------
#
#  WPS Complex Data formats
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from .codecs import CodecBase64, CodecRaw


class Format(object):
    """ Base complex data format.

    Constructor parameters:
        encoder     format's encoder object (defines the encoding)
        mime_type   mime-type of the format
        schema      optional schema of the document
        is_text     optional boolean flag indicating text-based data
                    format.
        is_xml      optional boolean flag indicating XML-based format.
                    The flag enables is_text flag.
        is_json     optional boolean flag indicating JSON-bases format.
                    The flag enables is_text flag.
    """
    # boolean flag indicating whether the format allows the payload to be
    # embedded to XML response or not. The XML embedding is disabled by default.
    allows_xml_embedding = False

    def __init__(self, encoder, mime_type, schema=None,
                 is_text=False, is_xml=False, is_json=False):
        # pylint: disable=too-many-arguments
        if is_xml or is_json:
            is_text = True
        self.mime_type = mime_type
        self.schema = schema
        self.is_text = is_text or is_xml or is_json
        self.is_xml = is_xml
        self.is_json = is_json
        self._codec = encoder

    @property
    def encoding(self):
        """ Get the format encoding name. """
        return self._codec.encoding

    def encode(self, file_in, **opt):
        """ Encoding generator."""
        return self._codec.encode(file_in, **opt)

    def decode(self, file_in, **opt):
        """ Encoding generator."""
        return self._codec.decode(file_in, **opt)


class FormatText(Format):
    """ Text-based complex data format. """
    allows_xml_embedding = True
    def __init__(self, mime_type="text/plain", schema=None,
                 text_encoding='utf-8'):
        Format.__init__(self, CodecRaw, mime_type, schema, True, False, False)
        self.text_encoding = text_encoding


class FormatXML(Format):
    """ XML-based complex data format. """
    allows_xml_embedding = True
    def __init__(self, mime_type="application/xml", schema=None,
                 text_encoding='utf-8'):
        Format.__init__(self, CodecRaw, mime_type, schema, True, True, False)
        self.text_encoding = text_encoding


class FormatJSON(Format):
    """ JSON-based complex data format. """
    allows_xml_embedding = True
    def __init__(self, mime_type="application/json", schema=None,
                 text_encoding='utf-8'):
        Format.__init__(self, CodecRaw, mime_type, schema, True, False, True)
        self.text_encoding = text_encoding


class FormatBinaryRaw(Format):
    """ Raw binary complex data format. """
    allows_xml_embedding = False
    def __init__(self, mime_type="application/octet-stream"):
        Format.__init__(self, CodecRaw, mime_type, None, False, False, False)


class FormatBinaryBase64(Format):
    """ Base64 encoded binary complex data format. """
    allows_xml_embedding = True
    def __init__(self, mime_type="application/octet-stream"):
        Format.__init__(self, CodecBase64, mime_type, None, False, False, False)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
from datetime import timedelta

from django.shortcuts import render
from django.utils.timezone import now
from django.db.models import Q, Max, Min

from eoxserver.core.util.timetools import isoformat
from eoxserver.resources.coverages import models


logger = logging.getLogger(__name__)


def index(request):
    return render(
        request, 'webclient/index.html', {
            "path": request.path,
        }
    )


def configuration(request):
    # select collections or coverages not contained in collections that are
    # visible
    qs = models.EOObject.objects.filter(
        Q(collection__isnull=False) |
        Q(mosaic__isnull=False) |
        Q(
            coverage__isnull=False,
            coverage__service_visibility__service="wc",
            coverage__service_visibility__visibility=True,
        )
    )

    # get the min/max values for begin and end time
    values = qs.aggregate(Min("begin_time"), Max("end_time"))
    start_time = values["begin_time__min"] or now() - timedelta(days=5)
    end_time = values["end_time__max"] or now()
    start_time_full = start_time - timedelta(days=5)
    end_time_full = end_time + timedelta(days=5)

    # try:
    #     # get only coverages that are in a collection or are visible
    #     # limit them to 10 and select the first time, so that we can limit the
    #     # initial brush
    #     coverages_qs = models.EOObject.objects.filter(
    #         Q(collection__isnull=True),
    #         Q(collections__isnull=False) | Q(coverage__visible=True)
    #     )
    #     first = list(coverages_qs.order_by("-begin_time")[:10])[-1]
    #     start_time = first.begin_time
    # except (models.EOObject.DoesNotExist, IndexError):
    #     pass

    return render(
        request, 'webclient/config.json', {
            "layers": qs,
            "start_time_full": isoformat(start_time_full),
            "end_time_full": isoformat(end_time_full),
            "start_time": isoformat(start_time),
            "end_time": isoformat(end_time)
        }
    )



# -----------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
# -----------------------------------------------------------------------------
# Copyright (C) 2012 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

"""
URLs config for EOxServer's {{ project_name }} instance.

"""
try:
    from django.conf.urls import include, url as re_path
except ImportError:
    from django.urls import include, re_path

from django.contrib import admin

from eoxserver.views import index


admin.autodiscover()


urlpatterns = [
    re_path(r'^$', index),
    re_path(r'^ows', include("eoxserver.services.urls")),
    re_path(r'^opensearch/', include('eoxserver.services.opensearch.urls')),

    # enable the client
    re_path(r'^client/', include('eoxserver.webclient.urls')),

    # Enable admin documentation:
    re_path(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    # Enable the admin:
    re_path(r'^admin/', admin.site.urls),
]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from logging import getLogger

from eoxserver.core.util import multiparttools as mp
from eoxserver.services.ows.wps.interfaces import (
    ProcessInterface, AsyncBackendInterface,
)
from eoxserver.services.ows.wps.util import (
    parse_named_parts, InMemoryURLResolver, get_processes
)
from eoxserver.services.ows.wps.exceptions import (
    NoSuchProcessError, InvalidParameterValue, StorageNotSupported,
)
from eoxserver.services.ows.wps.v10.encoders import (
    WPS10ExecuteResponseXMLEncoder, WPS10ExecuteResponseRawEncoder
)
from eoxserver.services.ows.wps.v10.execute_decoder_xml import (
    WPS10ExecuteXMLDecoder,
)
from eoxserver.services.ows.wps.v10.execute_decoder_kvp import (
    WPS10ExecuteKVPDecoder, parse_query_string,
)
from eoxserver.services.ows.wps.v10.execute_util import (
    parse_params, check_invalid_inputs, check_invalid_outputs,
    decode_raw_inputs, decode_output_requests, pack_outputs,
    resolve_request_parameters,
)

from eoxserver.services.ows.wps.util import get_async_backends


class WPS10ExecuteHandler(object):
    """ WPS 1.0 Execute service handler. """

    service = "WPS"
    versions = ("1.0.0",)
    request = "Execute"
    methods = ['GET', 'POST']

    @staticmethod
    def get_decoder(request):
        """ Get request decoder matching the request format. """
        if request.method == "GET":
            return WPS10ExecuteKVPDecoder(
                parse_query_string(request.META['QUERY_STRING'])
            )
        elif request.method == "POST":
            # support for multipart items
            if request.META["CONTENT_TYPE"].startswith("multipart/"):
                _, data = next(mp.iterate(request.body))
                return WPS10ExecuteXMLDecoder(data)
            return WPS10ExecuteXMLDecoder(request.body)

    def get_process(self, identifier):
        """ Get process component matched by the identifier. """
        for process in get_processes():
            process_identifier = (
                getattr(process, 'identifier', None) or type(process).__name__
            )
            if process_identifier == identifier:
                return process
        raise NoSuchProcessError(identifier)

    def get_async_backend(self):
        """ Get available asynchronous back-end matched by the service version.
        """
        version_set = set(self.versions)
        for backend in get_async_backends():
            if set(backend.supported_versions) & version_set:
                return backend

    def handle(self, request):
        # pylint: disable=redefined-variable-type, too-many-locals
        """ Request handler. """
        logger = getLogger(__name__)
        # decode request
        decoder = self.get_decoder(request)

        # parse named requests parts used in case of cid references
        extra_parts = parse_named_parts(request)

        # get the process and convert input/output definitions to a common format
        process = self.get_process(decoder.identifier)
        logger.debug("Execute process %s", decoder.identifier)
        input_defs = parse_params(process.inputs)
        output_defs = parse_params(process.outputs)

        # get the unparsed (raw) inputs and the requested response parameters
        raw_inputs = check_invalid_inputs(decoder.inputs, input_defs)
        resp_form = check_invalid_outputs(decoder.response_form, output_defs)

        # resolve the special request input parameters
        raw_inputs = resolve_request_parameters(raw_inputs, input_defs, request)

        if resp_form.raw:
            encoder = WPS10ExecuteResponseRawEncoder(resp_form)
        else:
            encoder = WPS10ExecuteResponseXMLEncoder(
                process, resp_form, raw_inputs
            )

        if not resp_form.raw and resp_form.store_response:
            # asynchronous execution
            async_backend = self.get_async_backend()
            if not async_backend:
                raise StorageNotSupported(
                    "This service instance does not support asynchronous "
                    "execution!"
                )

            if not getattr(process, 'asynchronous', False):
                raise StorageNotSupported(
                    "This process does not allow asynchronous execution!",
                )

            if not resp_form.status:
                raise InvalidParameterValue(
                    "The status update cannot be blocked for an asynchronous "
                    "execute request!", "status"
                )

            # pass the control over the processing to the asynchronous back-end
            job_id = async_backend.execute(
                process, raw_inputs, resp_form, extra_parts, request=request,
            )

            # encode the StatusAccepted response
            encoder.status_location = async_backend.get_response_url(job_id)
            response = encoder.encode_accepted()

        else:
            # synchronous execution
            if not getattr(process, 'synchronous', True):
                raise InvalidParameterValue(
                    "This process does not allow synchronous execution!",
                    "storeExecuteResponse"
                )

            if resp_form.status:
                raise InvalidParameterValue(
                    "The status update cannot be provided for a synchronous "
                    "execute request!", "status"
                )

            # prepare inputs passed to the process execution subroutine
            inputs = {}
            inputs.update(decode_output_requests(resp_form, output_defs))
            inputs.update(decode_raw_inputs(
                raw_inputs, input_defs, InMemoryURLResolver(extra_parts, logger)
            ))
            if not resp_form.raw:
                encoder.inputs = inputs

            # execute the process
            outputs = process.execute(**inputs)

            # pack the outputs
            packed_outputs = pack_outputs(outputs, resp_form, output_defs)

            # encode the StatusSucceeded response
            response = encoder.encode_response(packed_outputs)

        return encoder.serialize(response)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


try:
    from io import StringIO 
except ImportError:
    from cStringIO import StringIO

try:
    from PIL import Image, ImageFont, ImageDraw
    HAS_PIL = True
except ImportError:
    HAS_PIL = False
from lxml.builder import ElementMaker

from eoxserver.core import Component, implements
from eoxserver.core.decoders import kvp, lower
from eoxserver.core.util.xmltools import XMLEncoder, NameSpace, NameSpaceMap
from eoxserver.services.ows.interfaces import ExceptionHandlerInterface

try:
    # Python 2
    xrange
except NameError:
    # Python 3, xrange is now named range
    xrange = range

class WMS13ExceptionHandler(Component):
    implements(ExceptionHandlerInterface)

    service = "WMS"
    versions = ("1.3.0", "1.3")
    request = None

    def get_encoder(self, request):
        decoder = WMS13Decoder(request.GET)
        exceptions = decoder.exceptions
        if exceptions in ("xml", "application/vnd.ogc.se_xml") or not HAS_PIL:
            return WMS13ExceptionXMLEncoder()
        elif exceptions in ("inimage", "blank"):
            return WMS13ExceptionImageEncoder(
                decoder.width, decoder.height, decoder.format, decoder.bgcolor,
                exceptions=="blank"
            )
        print (decoder.exceptions)

    def handle_exception(self, request, exception):
        encoder = self.get_encoder(request)

        locator = getattr(exception, "locator", None)
        code = getattr(exception, "code", None) or type(exception).__name__

        return (
            encoder.serialize(
                encoder.encode_exception(
                    str(exception), code, locator
                ),
            ), 
            encoder.content_type, 
            400
        )


class WMS13Decoder(kvp.Decoder):
    width = kvp.Parameter(type=int, num="?")
    height = kvp.Parameter(type=int, num="?")
    format = kvp.Parameter(num="?")
    bgcolor = kvp.Parameter(num="?")
    exceptions = kvp.Parameter(num="?", type=lower, default="xml")


ns_ogc = NameSpace("http://www.opengis.net/ogc", "ogc")
nsmap = NameSpaceMap(ns_ogc)
OGC = ElementMaker(namespace=ns_ogc.uri, nsmap=nsmap)

class WMS13ExceptionXMLEncoder(XMLEncoder):
    def encode_exception(self, message, code, locator=None):
        attributes = {
            "code": code
        }
        if locator:
            attributes["locator"] = locator

        return OGC("ServiceExceptionReport",
            OGC("ServiceException",
                str(message),
                **attributes
            ),
            version="1.3.0"
        )

    @property
    def content_type(self):
        return "application/vnd.ogc.se_xml"

    def get_schema_locations(self):
        return {
            "http://www.opengis.net/ogc": "http://schemas.opengis.net/wms/1.3.0/exceptions_1_3_0.xsd"
        }


class WMS13ExceptionImageEncoder(object):
    def __init__(self, width=None, height=None, format=None, bgcolor=None, blank=False):
        self.width = width if width > 0 else 256
        self.height = height if height > 0 else 256
        if "/" in format:
            format = format[format.find("/") + 1:]
        self.format = format or "jpeg"
        self.bgcolor = bgcolor or "white"
        self.blank = blank

    @property
    def content_type(self):
        return "image/%s" % self.format

    def encode_exception(self, message, code, locator=None):
        width, height = self.width, self.height
        image = Image.new("RGB", (width, height), self.bgcolor)

        # if requested draw the exception string in the image
        if not self.blank:
            font = ImageFont.load_default()
            draw = ImageDraw.Draw(image)
            yoffset = 0
            while len(message):
                for i in xrange(len(message)):
                    part = message if i == 0 else message[:-i]
                    xsize, ysize = font.getsize(part)
                    print (i, xsize, ysize, part)
                    if xsize < width:
                        break
                draw.text((0, yoffset), part, font=font, fill="red")
                yoffset += ysize
                message = message[-i:]
                if i == 0:
                    break
        return image

    def serialize(self, image):
        f = StringIO()
        try:
            image.save(f, self.format)
        except (IOError, KeyError):
            image.save(f, "jpeg") # Fallback solution
        return f.getvalue()



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging
from itertools import chain

from django.utils.datastructures import SortedDict

from eoxserver.core import Component, ExtensionPoint
from eoxserver.core.config import get_eoxserver_config
from eoxserver.contrib import mapserver as ms
from eoxserver.resources.coverages.crss import CRSsConfigReader
from eoxserver.services.mapserver.interfaces import (
    ConnectorInterface, LayerFactoryInterface, StyleApplicatorInterface
)
from eoxserver.services.result import result_set_from_raw_data, get_content_type
from eoxserver.services.exceptions import RenderException
from eoxserver.services.ows.wms.exceptions import InvalidFormat

logger = logging.getLogger(__name__)


class MapServerWMSBaseComponent(Component):
    """ Base class for various WMS render components using MapServer.
    """

    connectors = ExtensionPoint(ConnectorInterface)
    layer_factories = ExtensionPoint(LayerFactoryInterface)
    style_applicators = ExtensionPoint(StyleApplicatorInterface)

    def render(self, layer_groups, request_values, **options):
        map_ = ms.Map()
        map_.setMetaData("ows_enable_request", "*")
        map_.setProjection("EPSG:4326")
        map_.imagecolor.setRGB(0, 0, 0)

        # set supported CRSs
        decoder = CRSsConfigReader(get_eoxserver_config())
        crss_string = " ".join(
            map(lambda crs: "EPSG:%d" % crs, decoder.supported_crss_wms)
        )
        map_.setMetaData("ows_srs", crss_string)
        map_.setMetaData("wms_srs", crss_string)

        self.check_parameters(map_, request_values)

        session = self.setup_map(layer_groups, map_, options)

        with session:
            request = ms.create_request(request_values)
            raw_result = map_.dispatch(request)

            result = result_set_from_raw_data(raw_result)
            return result, get_content_type(result)

    def check_parameters(self, map_, request_values):
        for key, value in request_values:
            if key.lower() == "format":
                if not map_.getOutputFormatByName(value):
                    raise InvalidFormat(value)
                break
        else:
            raise RenderException("Missing 'format' parameter")

    @property
    def suffixes(self):
        return list(
            chain(*[factory.suffixes for factory in self.layer_factories])
        )

    def get_connector(self, data_items):
        for connector in self.connectors:
            if connector.supports(data_items):
                return connector
        return None

    def get_layer_factory(self, suffix):
        result = None
        for factory in self.layer_factories:
            if suffix in factory.suffixes:
                if result:
                    pass  # TODO
                    #raise Exception("Found")
                result = factory
                return result
        return result

    def setup_map(self, layer_selection, map_, options):
        group_layers = SortedDict()
        session = ConnectorSession(options)

        # set up group layers before any "real" layers
        for collections, _, name, suffix in tuple(layer_selection.walk()):
            if not collections:
                continue

            # get a factory for the given suffix
            factory = self.get_layer_factory(suffix)
            if not factory:
                # raise or pass?
                continue

            # get the groups name, which is the name of the collection + the
            # suffix
            group_name = collections[-1].identifier + (suffix or "")

            # generate a group layer
            group_layer = factory.generate_group(group_name)
            group_layers[group_name] = group_layer

        # set up the actual layers for each coverage
        for collections, coverage, name, suffix in layer_selection.walk():
            # get a factory for the given coverage and suffix
            factory = self.get_layer_factory(suffix)

            group_layer = None
            group_name = None

            if collections:
                group_name = collections[-1].identifier + (suffix or "")
                group_layer = group_layers.get(group_name)

            if not coverage:
                # add an empty layer to not produce errors out of bounds.
                if name:
                    tmp_layer = ms.layerObj()
                    tmp_layer.name = (name + suffix) if suffix else name
                    layers_and_data_items = ((tmp_layer, ()),)
                else:
                    layers_and_data_items = ()

            elif not factory:
                tmp_layer = ms.layerObj()
                tmp_layer.name = name
                layers_and_data_items = ((tmp_layer, ()),)
            else:
                data_items = coverage.data_items.all()
                coverage.cached_data_items = data_items
                layers_and_data_items = tuple(factory.generate(
                    coverage, group_layer, suffix, options
                ))

            for layer, data_items in layers_and_data_items:
                connector = self.get_connector(data_items)

                if group_name:
                    layer.setMetaData("wms_layer_group", "/" + group_name)

                session.add(connector, coverage, data_items, layer)

        coverage_layers = [layer for _, layer, _ in session.coverage_layers]
        for layer in chain(group_layers.values(), coverage_layers):
            old_layer = map_.getLayerByName(layer.name)
            if old_layer:
                # remove the old layer and reinsert the new one, to
                # raise the layer to the top.
                # TODO: find a more efficient way to do this
                map_.removeLayer(old_layer.index)
            map_.insertLayer(layer)

        # apply any styles
        # TODO: move this to map/legendgraphic renderer only?
        for coverage, layer, data_items in session.coverage_layers:
            for applicator in self.style_applicators:
                applicator.apply(coverage, data_items, layer)

        return session

    def get_empty_layers(self, name):
        layer = ms.layerObj()
        layer.name = name
        layer.setMetaData("wms_enable_request", "getmap")
        return (layer,)


class ConnectorSession(object):
    """ Helper class to be used in `with` statements. Allows connecting and
        disconnecting all added layers with the given data items.
    """
    def __init__(self, options=None):
        self.item_list = []
        self.options = options or {}

    def add(self, connector, coverage, data_items, layer):
        self.item_list.append(
            (connector, coverage, layer, data_items)
        )

    def __enter__(self):
        for connector, coverage, layer, data_items in self.item_list:
            if connector:
                connector.connect(coverage, data_items, layer, self.options)

    def __exit__(self, *args, **kwargs):
        for connector, coverage, layer, data_items in self.item_list:
            if connector:
                connector.disconnect(coverage, data_items, layer, self.options)

    @property
    def coverage_layers(self):
        return map(lambda it: (it[1], it[2], it[3]), self.item_list)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.db.models import Case, Value, When, IntegerField, BooleanField

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, enum
from eoxserver.core.util.timetools import isoformat
from eoxserver.render.map.objects import (
    CoverageLayer, CoveragesLayer, OutlinedCoveragesLayer, MosaicLayer,
    OutlinesLayer, BrowseLayer, OutlinedBrowseLayer,
    MaskLayer, MaskedBrowseLayer,
    LayerDescription,
)
from eoxserver.render.coverage.objects import Coverage as RenderCoverage
from eoxserver.render.coverage.objects import Mosaic as RenderMosaic
from eoxserver.render.browse.objects import (
    Browse, GeneratedBrowse, Mask, MaskedBrowse
)
from eoxserver.resources.coverages import models


class UnsupportedObject(Exception):
    pass


class NoSuchLayer(Exception):
    code = 'LayerNotDefined'
    locator = 'layer'


class NoSuchPrefix(NoSuchLayer):
    code = 'LayerNotDefined'
    locator = 'layer'


class LayerMapper(object):
    """ Default layer mapper.
    """

    def __init__(self, supported_layer_types, suffix_separator):
        self.supported_layer_types = supported_layer_types
        self.suffix_separator = suffix_separator

    def get_layer_description(self, eo_object, raster_styles, geometry_styles):
        if isinstance(eo_object, models.Coverage):
            coverage = RenderCoverage.from_model(eo_object)
            return LayerDescription.from_coverage(coverage, raster_styles)
        elif isinstance(eo_object, models.Mosaic):
            coverage = RenderCoverage.from_model(eo_object)
            return LayerDescription.from_mosaic(coverage, raster_styles)
        elif isinstance(eo_object, (models.Product, models.Collection)):
            if getattr(eo_object, "product_type", None):
                browse_type_qs = eo_object.product_type.browse_types.all()
                mask_type_qs = eo_object.product_type.mask_types.all()
            elif getattr(eo_object, "collection_type", None):
                browse_type_qs = models.BrowseType.objects.filter(
                    product_type__allowed_collection_types__collections=eo_object
                )
                mask_type_qs = models.MaskType.objects.filter(
                    product_type__allowed_collection_types__collections=eo_object
                )
            else:
                browse_type_qs = models.BrowseType.objects.none()
                mask_type_qs = models.MaskType.objects.none()

            browse_types_name_and_is_gray = browse_type_qs.annotate(
                is_gray=Case(
                    When(green_expression__isnull=True, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField()
                )
            ).values_list('name', 'is_gray').distinct()
            mask_type_names = mask_type_qs.values_list(
                'name', flat=True
            ).distinct()

            sub_layers = [
                LayerDescription(
                    "%s%soutlines" % (
                        eo_object.identifier, self.suffix_separator
                    ),
                    styles=geometry_styles,
                    queryable=True
                ),
                LayerDescription(
                    "%s%soutlined" % (
                        eo_object.identifier, self.suffix_separator
                    ),
                    styles=geometry_styles,
                    queryable=True
                )
            ]
            for name, is_gray in browse_types_name_and_is_gray:
                sub_layers.append(
                    LayerDescription(
                        "%s%s%s" % (
                            eo_object.identifier, self.suffix_separator,
                            name
                        ) if name else eo_object.identifier,
                        styles=raster_styles if is_gray else []
                    )
                )

            for mask_type_name in mask_type_names:
                sub_layers.append(
                    LayerDescription(
                        "%s%s%s" % (
                            eo_object.identifier, self.suffix_separator,
                            mask_type_name
                        ),
                        styles=geometry_styles
                    )
                )
                sub_layers.append(
                    LayerDescription(
                        "%s%smasked_%s" % (
                            eo_object.identifier, self.suffix_separator,
                            mask_type_name
                        )
                    )
                )

            dimensions = {}
            if eo_object.begin_time and eo_object.end_time:
                dimensions["time"] = {
                    'min': isoformat(eo_object.begin_time),
                    'max': isoformat(eo_object.end_time),
                    'step': 'PT1S',
                    'default': isoformat(eo_object.end_time),
                    'units': 'ISO8601'
                }

            return LayerDescription(
                name=eo_object.identifier,
                bbox=eo_object.footprint.extent if eo_object.footprint else None,
                dimensions=dimensions,
                sub_layers=sub_layers
            )

        raise UnsupportedObject(
            "Object %r cannot be mapped to a layer." % eo_object
        )

    def lookup_layer(self, layer_name, suffix, style, filters_expressions,
                     sort_by, time, ranges, bands, wavelengths, elevation,
                     zoom):
        """ Lookup the layer from the registered objects.
        """
        reader = LayerMapperConfigReader(get_eoxserver_config())
        limit_products = (
            reader.limit_products if reader.limit_mode == 'hide' else None
        )
        min_render_zoom = reader.min_render_zoom
        full_name = '%s%s%s' % (layer_name, self.suffix_separator, suffix)

        try:
            eo_object = models.EOObject.objects.select_subclasses(
                models.Collection, models.Product, models.Coverage,
                models.Mosaic
            ).get(
                identifier=layer_name
            )
        except models.EOObject.DoesNotExist:
            raise NoSuchLayer('Layer %r does not exist' % layer_name)

        if isinstance(eo_object, models.Coverage):
            if suffix in ('', 'bands'):
                return CoverageLayer(
                    full_name, style,
                    RenderCoverage.from_model(eo_object),
                    bands, wavelengths, time, elevation, ranges
                )

            elif suffix == 'outlines':
                return OutlinesLayer(
                    name=full_name, style=style, fill=None,
                    footprints=[eo_object.footprint]
                )

            # TODO: masked coverages, when using the coverages product

            else:
                raise NoSuchLayer('Invalid layer suffix %r' % suffix)

        # TODO: deprecated
        elif isinstance(eo_object, models.Mosaic):
            if suffix == 'outlines':
                return OutlinesLayer(
                    name=full_name, style=style, fill=None,
                    footprints=[
                        coverage.footprint
                        for coverage in self.iter_coverages(
                            eo_object, filters_expressions, sort_by
                        )
                    ]
                )
            else:
                return MosaicLayer(
                    full_name, style,
                    RenderMosaic.from_model(eo_object), [
                        RenderCoverage.from_model(coverage)
                        for coverage in self.iter_coverages(
                            eo_object, filters_expressions, sort_by
                        )
                    ], bands, wavelengths, time, elevation, ranges
                )

        elif isinstance(eo_object, (models.Collection, models.Product)):
            if suffix == '' or suffix == 'outlined' or suffix == 'bands':
                browses = []
                product_browses = self.iter_products_browses(
                    eo_object, filters_expressions, sort_by, None, style,
                    limit=limit_products
                )

                has_products = False
                for product, browse, _ in product_browses:
                    # When bands/wavelengths are specifically requested, make a
                    # generated browse
                    if bands or wavelengths:
                        browse = _generate_browse_from_bands(
                            product, bands, wavelengths, ranges
                        )
                        if browse:
                            browses.append(browse)
                            has_products = True

                    # When available use the default browse
                    elif browse:
                        browses.append(Browse.from_model(product, browse))
                        has_products = True

                    # As fallback use the default browse type (with empty name)
                    # to generate a browse from the specified bands
                    else:
                        browse_type = product.product_type.browse_types.filter(
                            name=''
                        ).first()
                        if browse_type:
                            browse = _generate_browse_from_browse_type(
                                product, browse_type
                            )
                            if browse:
                                browses.append(browse)
                                has_products = True

                if not has_products:
                    coverages = self.iter_coverages(
                        eo_object, filters_expressions, sort_by
                    )

                    if suffix == '':
                        return CoveragesLayer(
                            full_name, style, [
                                RenderCoverage.from_model(coverage)
                                for coverage in coverages
                            ],
                            bands, wavelengths, time, elevation, ranges
                        )
                    else:
                        return OutlinedCoveragesLayer(
                            full_name, style, [
                                RenderCoverage.from_model(coverage)
                                for coverage in coverages
                            ],
                            bands, wavelengths, time, elevation, ranges
                        )

                # detect whether we are below the zoom limit
                if min_render_zoom is None or zoom >= min_render_zoom:
                    # either return the simple browse layer or the outlined one
                    if suffix == '':
                        return BrowseLayer(
                            name=full_name, style=style,
                            browses=browses, ranges=ranges
                        )
                    else:
                        return OutlinedBrowseLayer(
                            name=full_name, style=style,
                            browses=browses, ranges=ranges
                        )

                # render outlines when we are below the zoom limit
                else:
                    return OutlinesLayer(
                        name=full_name, style=reader.color,
                        fill=reader.fill_opacity,
                        footprints=[
                            product.footprint for product in self.iter_products(
                                eo_object, filters_expressions, sort_by,
                                limit=limit_products
                            )
                        ]
                    )

            elif suffix == 'outlines':
                return OutlinesLayer(
                    name=full_name, style=style, fill=None,
                    footprints=[
                        product.footprint for product in self.iter_products(
                            eo_object, filters_expressions, sort_by,
                            limit=limit_products
                        )
                    ]
                )

            elif suffix.startswith('outlines_masked_'):
                post_suffix = suffix[len('outlines_masked_'):]

                product_browses_mask = self.iter_products_browses_masks(
                    eo_object, filters_expressions, sort_by, post_suffix,
                    limit=limit_products
                )
                footprints = []
                masks = []
                for product, browse, mask, mask_type in product_browses_mask:
                    footprints.append(product.footprint)
                    masks.append(Mask.from_model(mask, mask_type))

                return OutlinesLayer(
                    name=full_name, style=style, fill=None,
                    footprints=footprints,
                    masks=masks,
                )

            elif suffix.startswith('masked_'):
                post_suffix = suffix[len('masked_'):]
                mask_type = self.get_mask_type(eo_object, post_suffix)

                if not mask_type:
                    raise NoSuchLayer('No such mask type %r' % post_suffix)

                masked_browses = []

                product_browses_mask = self.iter_products_browses_masks(
                    eo_object, filters_expressions, sort_by, post_suffix,
                    limit=limit_products
                )
                for product, browse, mask, mask_type in product_browses_mask:
                    # When bands/wavelengths are specifically requested, make a
                    # generated browse
                    if bands or wavelengths:
                        masked_browses.append(
                            MaskedBrowse(
                                browse=_generate_browse_from_bands(
                                    product, bands, wavelengths, ranges
                                ),
                                mask=Mask.from_model(mask, mask_type)
                            )
                        )

                    # When available use the default browse
                    elif browse:
                        masked_browses.append(
                            MaskedBrowse.from_models(
                                product, browse, mask, mask_type
                            )
                        )

                    # As fallback use the default browse type (with empty name)
                    # to generate a browse from the specified bands
                    else:
                        browse_type = product.product_type.browse_types.filter(
                            name=''
                        ).first()
                        if browse_type:
                            masked_browses.append(
                                MaskedBrowse(
                                    browse=_generate_browse_from_browse_type(
                                        product, browse_type
                                    ),
                                    mask=Mask.from_model(mask, mask_type)
                                )
                            )

                return MaskedBrowseLayer(
                    name=full_name, style=style,
                    masked_browses=masked_browses
                )

            else:
                # either browse type or mask type
                browse_type = self.get_browse_type(eo_object, suffix)
                if browse_type:
                    browses = []

                    product_browses = self.iter_products_browses(
                        eo_object, filters_expressions, sort_by, suffix,
                        style, limit=limit_products
                    )

                    for product, browse, browse_type in product_browses:
                        # check if a browse is already available for that
                        # browse type.
                        if browse:
                            browses.append(Browse.from_model(product, browse))

                        # if no browse is available for that browse type,
                        # generate a new browse with the instructions of that
                        # browse type
                        else:
                            browse = _generate_browse_from_browse_type(
                                product, browse_type
                            )
                            if browse:
                                browses.append(browse)

                    return BrowseLayer(
                        name=full_name, style=style, ranges=ranges,
                        browses=browses
                    )

                mask_type = self.get_mask_type(eo_object, suffix)
                if mask_type:
                    return MaskLayer(
                        name=full_name, style=style,
                        masks=[
                            Mask.from_model(mask_model, mask_type)
                            for _, mask_model in self.iter_products_masks(
                                eo_object, filters_expressions, sort_by, suffix,
                                limit=limit_products
                            )
                        ]
                    )

                raise NoSuchLayer('Invalid layer suffix %r' % suffix)

    def split_layer_suffix_name(self, layer_name):
        return layer_name.partition(self.suffix_separator)[::2]

    def get_browse_type(self, eo_object, name):
        if isinstance(eo_object, models.Product):
            filter_ = dict(product_type__products=eo_object)
        else:
            filter_ = dict(
                product_type__allowed_collection_types__collections=eo_object
            )

        return models.BrowseType.objects.filter(name=name, **filter_).first()

    def get_mask_type(self, eo_object, name):
        if isinstance(eo_object, models.Product):
            filter_ = dict(product_type__products=eo_object)
        else:
            filter_ = dict(
                product_type__allowed_collection_types__collections=eo_object
            )

        return models.MaskType.objects.filter(name=name, **filter_).first()

    #
    # iteration methods
    #

    def iter_coverages(self, eo_object, filters_expressions, sort_by=None):
        if isinstance(eo_object, models.Mosaic):
            base_filter = dict(mosaics=eo_object)
        elif isinstance(eo_object, models.Collection):
            base_filter = dict(collections=eo_object)
        elif isinstance(eo_object, models.Product):
            base_filter = dict(parent_product=eo_object)

        qs = models.Coverage.objects.filter(filters_expressions, **base_filter)
        if sort_by:
            qs = qs.order_by('%s%s' % (
                '-' if sort_by[1] == 'DESC' else '',
                sort_by[0]
            ))
        else:
            qs = qs.order_by(
                '-begin_time', '-end_time', 'identifier'
            )

        return qs

    def iter_products(self, eo_object, filters_expressions, sort_by=None,
                      limit=None):
        if isinstance(eo_object, models.Collection):
            base_filter = dict(collections=eo_object)
        else:
            base_filter = dict(pk=eo_object.pk)

        qs = models.Product.objects.filter(filters_expressions, **base_filter)
        if limit is not None:
            qs = qs[:limit]

        if sort_by:
            qs = qs.order_by('%s%s' % (
                '-' if sort_by[1] == 'DESC' else '',
                sort_by[0]
            ))
        else:
            qs = qs.order_by(
                '-begin_time', '-end_time', 'identifier'
            )

        return qs

    def iter_products_browses(self, eo_object, filters_expressions, sort_by,
                              name=None, style=None, limit=None):
        products = self.iter_products(
            eo_object, filters_expressions, sort_by, limit
        ).prefetch_related('browses')

        for product in products:
            browses = product.browses
            if name:
                browses = browses.filter(browse_type__name=name)
                browse = browses.first()
                if browse:
                    browse_type = browse.browse_type
                else:
                    browse_type = models.BrowseType.objects.get(
                        name=name, product_type=product.product_type
                    )
            else:
                browses = browses.filter(browse_type__isnull=True)
                browse_type = None

            # if style:
            #     browses = browses.filter(style=style)
            # else:
            #     browses = browses.filter(style__isnull=True)

            yield (product, browses.first(), browse_type)

    def iter_products_masks(self, eo_object, filters_expressions, sort_by,
                            name=None, limit=None):
        products = self.iter_products(
            eo_object, filters_expressions, sort_by, limit
        ).prefetch_related('masks')

        for product in products:
            masks = product.masks
            if name:
                mask = masks.filter(mask_type__name=name).first()
            else:
                mask = masks.filter(mask_type__isnull=True).first()

            yield (product, mask)

    def iter_products_browses_masks(self, eo_object, filters_expressions,
                                    sort_by, name=None, limit=None):
        products = self.iter_products(
            eo_object, filters_expressions, sort_by, limit
        ).prefetch_related('masks', 'browses')

        for product in products:
            if name:
                mask = product.masks.filter(mask_type__name=name).first()
                mask_type = models.MaskType.objects.filter(
                    product_type=product.product_type, name=name
                ).first()
            else:
                mask = product.masks.filter(mask_type__isnull=True).first()
                mask_type = None

            browse = product.browses.filter(browse_type__isnull=True).first()

            yield (product, browse, mask, mask_type)


class LayerMapperConfigReader(config.Reader):
    section = "services.ows.wms"
    limit_products = config.Option(type=int)
    limit_mode = config.Option(type=enum('hide', 'outlines'), default='hide')
    min_render_zoom = config.Option(type=int)
    fill_opacity = config.Option(type=float)
    color = config.Option(type=str, default='grey')


def _generate_browse_from_browse_type(product, browse_type):
    if not browse_type.red_or_grey_expression:
        return None

    from eoxserver.render.browse.generate import extract_fields

    band_expressions = []
    field_names = []
    red_bands = extract_fields(browse_type.red_or_grey_expression)
    band_expressions.append((
        browse_type.red_or_grey_expression, (
            browse_type.red_or_grey_range_min,
            browse_type.red_or_grey_range_max,
        )
    ))
    field_names.extend(red_bands)

    if browse_type.green_expression and browse_type.blue_expression:
        green_bands = extract_fields(browse_type.green_expression)
        blue_bands = extract_fields(browse_type.blue_expression)
        band_expressions.append((
            browse_type.green_expression, (
                browse_type.green_range_min,
                browse_type.green_range_max,
            )
        ))
        band_expressions.append((
            browse_type.blue_expression, (
                browse_type.blue_range_min,
                browse_type.blue_range_max,
            )
        ))
        field_names.extend(green_bands)
        field_names.extend(blue_bands)

        if browse_type.alpha_expression:
            alpha_bands = extract_fields(browse_type.alpha_expression)
            band_expressions.append((
                browse_type.alpha_expression, (
                    browse_type.alpha_range_min,
                    browse_type.alpha_range_max,
                )
            ))
            field_names.extend(alpha_bands)

    coverages, fields_and_coverages = _lookup_coverages(product, field_names)

    # only return a browse instance if coverages were found
    if coverages:
        return GeneratedBrowse.from_coverage_models(
            band_expressions, fields_and_coverages, product
        )
    return None


def _generate_browse_from_bands(product, bands, wavelengths, ranges):
    assert len(bands or wavelengths or []) in (1, 3, 4)
    if bands:
        coverages, fields_and_coverages = _lookup_coverages(product, bands)
    # TODO: implement with wavelengths
    # elif wavelengths:
    #     fields_and_coverages = [
    #         (
    #             [product.coverages.filter(
    #                 coverage_type__field_types__wavelength=wavelength
    #             ).first().name],
    #             product.coverages.filter(
    #                 coverage_type__field_types__wavelength=wavelength
    #             )
    #         )
    #         for wavelength in wavelengths
    #     ]
    # only return a browse instance if coverages were found
    if coverages:
        return GeneratedBrowse.from_coverage_models(
            zip(bands, ranges or [(None, None)] * len(bands)), fields_and_coverages, product
        )
    return None


def _lookup_coverages(product, field_names):
    # make a query of all coverages in that product for the given fields
    coverages = product.coverages.filter(
        coverage_type__field_types__identifier__in=field_names
    )

    # annotate the coverages with booleans indicating whether or not they have a
    # certain field
    coverages = coverages.annotate(**{
        'has_%s' % field_name: Case(
            When(
                coverage_type__field_types__identifier=field_name,
                then=Value(1)
            ),
            default=Value(0),
            output_field=IntegerField()
        )
        for field_name in field_names
    })

    # evaluate the queryset
    coverages = list(coverages)

    # make a dictionary for all field mapping to their respective coverages
    fields_and_coverages = {
        field_name: [
            coverage
            for coverage in coverages
            if getattr(coverage, 'has_%s' % field_name)
        ]
        for field_name in field_names
    }
    return coverages, fields_and_coverages



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.contrib.gis import admin

from eoxserver.resources.coverages.admin import (
    CollectionAdmin, ProductAdmin, CoverageAdmin
)
from eoxserver.services import models


class ServiceVisibilityInline(admin.TabularInline):
    model = models.ServiceVisibility
    extra = 0


# register inline

for admin_class in [CollectionAdmin, ProductAdmin, CoverageAdmin]:
    if admin_class in admin.site._registry:
        admin.site._registry[admin_class].inlines.append(ServiceVisibilityInline)
    else:
        admin_class.inlines.append(ServiceVisibilityInline)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage grids. This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')

        for parser in [create_parser, delete_parser]:
            parser.add_argument(
                'name', nargs=1, help='The grid name'
            )

        create_parser.add_argument(
            'coordinate_reference_system', nargs=1,
            help=(
                'The definition of the coordinate reference system. Either '
                'an integer (the EPSG code), or the URL, WKT or XML definiton.'
            )
        )

        create_parser.add_argument(
            '--name', '--axis-name', '-n', dest='axis_names', default=[],
            action='append',
            help=(
                'The name of one axis. Must be passed at least once and up to '
                'four times.'
            )
        )
        create_parser.add_argument(
            '--type', '--axis-type', '-t', dest='axis_types', default=[],
            action='append',
            choices=[choice[1] for choice in models.Grid.AXIS_TYPES],
            help=(
                'The type of one axis. Must be passed at least once and up to '
                'four times.'
            )
        )
        create_parser.add_argument(
            '--offset', '--axis-offset', '-o', dest='axis_offsets', default=[],
            action='append',
            help=(
                'The offset for one axis. Must be passed at least once and up '
                'to four times.'
            )
        )

    @transaction.atomic
    def handle(self, subcommand, name, *args, **kwargs):
        """ Dispatch sub-commands: create, delete.
        """
        name = name[0]
        if subcommand == "create":
            self.handle_create(name, *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(name, *args, **kwargs)

    def handle_create(self, name, coordinate_reference_system, **kwargs):
        """ Handle the creation of a new product
        """
        axis_names = kwargs['axis_names']
        axis_types = kwargs['axis_types']
        axis_offsets = kwargs['axis_offsets']

        if not axis_names:
            raise CommandError('Must supply at least one axis definition.')

        if len(axis_types) != len(axis_names):
            raise CommandError(
                'Invalid number of axis-types supplied. Expected %d, got %d.'
                % (len(axis_names), len(axis_types))
            )
        if len(axis_offsets) != len(axis_names):
            raise CommandError(
                'Invalid number of axis-offsets supplied. Expected %d, got %d.'
                % (len(axis_names), len(axis_offsets))
            )

        if len(axis_names) > 4:
            raise CommandError('Currently only at most four axes are supported.')

        type_name_to_id = dict(
            (name, id_) for id_, name in models.Grid.AXIS_TYPES
        )

        iterator = enumerate(zip(axis_names, axis_types, axis_offsets), start=1)
        definition = {
            'name': name,
            'coordinate_reference_system': coordinate_reference_system[0]
        }
        for i, (name, type_, offset) in iterator:
            definition['axis_%d_name' % i] = name
            definition['axis_%d_type' % i] = type_name_to_id[type_]
            definition['axis_%d_offset' % i] = offset

        models.Grid.objects.create(**definition)

    def handle_delete(self, name, **kwargs):
        """ Handle the deregistration a product
        """
        try:
            models.Grid.objects.get(name=name).delete()
        except models.Grid.DoesNotExist:
            raise CommandError('No such Grid %r' % name)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging
import itertools
from functools import partial

from django.http import HttpResponse
from django.utils.six import string_types

from eoxserver.core import Component, ExtensionPoint, env
from eoxserver.services.ows.interfaces import *
from eoxserver.services.ows.decoders import get_decoder
from eoxserver.services.exceptions import (
    ServiceNotSupportedException, VersionNotSupportedException,
    VersionNegotiationException, OperationNotSupportedException,
    HTTPMethodNotAllowedError,
)
from eoxserver.services.ows.common.v20.exceptionhandler import (
    OWS20ExceptionHandler
)


logger = logging.getLogger(__name__)

ALLOWED_HTTP_METHODS = ["GET", "POST", "OPTIONS"]

class OptionsRequestHandler(object):
    """ Dummy request handler class to respond to HTTP OPTIONS requests.
    """
    def handle(self, request):

        def add_required_headers(headers, required_headers):
            """ Make sure the required headers are included in the list. """
            headers_lc = set(header.lower() for header in headers)
            for required_header in required_headers:
                if required_header.lower() not in headers_lc:
                    headers.append(required_header)
            return headers

        # return an empty 200 response
        response = HttpResponse()
        response["Access-Control-Allow-Methods"] = ", ".join(
            ALLOWED_HTTP_METHODS
        )
        headers = [
            header.strip() for header in
            request.META.get("HTTP_ACCESS_CONTROL_REQUEST_HEADERS", "").split(",")
            if header
        ]
        headers = add_required_headers(headers, ['Content-Type'])
        response["Access-Control-Allow-Headers"] = ", ".join(headers)
        return response


class ServiceComponent(Component):
    service_handlers = ExtensionPoint(ServiceHandlerInterface)
    exception_handlers = ExtensionPoint(ExceptionHandlerInterface)

    get_service_handlers = ExtensionPoint(GetServiceHandlerInterface)
    post_service_handlers = ExtensionPoint(PostServiceHandlerInterface)

    version_negotiation_handlers = ExtensionPoint(VersionNegotiationInterface)

    def __init__(self, *args, **kwargs):
        super(ServiceComponent, self).__init__(*args, **kwargs)

    def query_service_handler(self, request):
        """ Tries to find the correct service handler for a given request. The
        request ``method`` can either be "POST" (in which case the request body
        is parsed as XML) or "GET" (in which case the request is parsed
        as "KVP").

        If necessary a version negotiation is conducted, following OWS
        guidelines.

        :param request: a :class:`Django HttpRequest <django.http.HttpRequest>`
                        object
        :returns: the request handler component for the given request
        :raises ServiceNotSupportedException: if the service is not supported
                                              by any component
        :raises VersionNotSupportedException: if the specified version is not
                                              supported
        :raises OperationNotSupportedException: if the specified request
                                                operation is not supported
        """

        decoder = get_decoder(request)


        if request.method == "GET":
            handlers = self.get_service_handlers
        elif request.method == "POST":
            handlers = self.post_service_handlers
        elif request.method == "OPTIONS":
            return OptionsRequestHandler()
        else:
            raise HTTPMethodNotAllowedError(
                "The %s HTTP method is not allowed!" % request.method,
                ALLOWED_HTTP_METHODS
            )
            #handlers = self.service_handlers

        version = decoder.version
        if version is None:
            accepted_versions = decoder.acceptversions
            handlers = filter_handlers(
                handlers, decoder.service, accepted_versions, decoder.request
            )
            return self.version_negotiation(handlers, accepted_versions)

        # check that the service is supported
        handlers = filter(
            partial(handler_supports_service, service=decoder.service), handlers
        )
        if not handlers:
            raise ServiceNotSupportedException(decoder.service)

        # check that the required version is enabled
        handlers_ = filter(
            lambda h: decoder.version in h.versions, handlers
        )
        if not handlers_:
            # old style version negotiation shall always return capabilities
            if decoder.request == "GETCAPABILITIES":
                handlers = [sorted(
                    filter(
                        lambda h: decoder.request == h.request.upper(), handlers
                    ), key=lambda h: max(h.versions), reverse=True
                )[0]]
            else:
                raise VersionNotSupportedException(
                    decoder.service, decoder.version
                )
        else:
            handlers = handlers_

        # check that the required operation is supported and sort by the highest
        # version supported in descending manner
        handlers = sorted(
            filter(
                lambda h: decoder.request == h.request.upper(), handlers
            ), key=lambda h: max(h.versions), reverse=True
        )

        if not handlers:
            operation = decoder.request
            raise OperationNotSupportedException(
                "Operation '%s' is not supported." % operation, operation
            )

        # return the handler with the highest version
        logger.debug("Handling '%s' request for '%s' service version '%s'." %
                     (handlers[0].request, handlers[0].service,
                      handlers[0].versions[0]))
        return handlers[0]

    def query_service_handlers(self, service=None, versions=None, request=None,
                               method=None):
        """ Query the service handler components, filtering optionally by
            ``service``, ``versions``, ``request`` or ``method``.
        """
        method = method.upper() if method is not None else None

        if method == "GET":
            handlers = self.get_service_handlers
        elif method == "POST":
            handlers = self.post_service_handlers
        elif method is None:
            handlers = self.service_handlers
        else:
            return []

        handlers = filter_handlers(handlers, service, versions, request)
        return sort_handlers(handlers)

    def query_exception_handler(self, request):
        try:
            decoder = get_decoder(request)
            handlers = self.exception_handlers
            handlers = sorted(
                filter(
                    partial(handler_supports_service, service=decoder.service),
                    self.exception_handlers
                ),
                key=lambda h: max(h.versions), reverse=True
            )

            # try to get the correctly versioned exception handler
            if decoder.version:
                for handler in handlers:
                    if decoder.version in handler.versions:
                        return handler
            else:
                # return the exception handler with the highest version,
                # if one is available
                return handlers[0]
        except:
            # swallow any exception here, because we *really* need a handler
            # to correctly show the exception.
            pass

        # last resort fallback is a plain OWS exception handler
        return OWS20ExceptionHandler()

    def version_negotiation(self, handlers, accepted_versions=None):
        version_to_handler = {}
        for handler in handlers:
            for version in handler.versions:
                version_to_handler.setdefault(version, handler)

        available_versions = sorted(version_to_handler.keys(), reverse=True)
        if not available_versions:
            raise VersionNegotiationException()

        if not accepted_versions:
            return version_to_handler[available_versions[0]]

        combinations = itertools.product(accepted_versions, available_versions)
        for accepted_version, available_version in combinations:
            if accepted_version == available_version:
                return version_to_handler[available_version]

        raise VersionNegotiationException()


def filter_handlers(handlers, service=None, versions=None, request=None):
    """ Utility function to filter the given OWS service handlers by their
        attributes 'service', 'versions' and 'request'.
    """

    service = service.upper() if service is not None else None
    request = request.upper() if request is not None else None

    if service:
        handlers = filter(
            partial(handler_supports_service, service=service), handlers
        )

    if request:
        handlers = filter(lambda h: h.request.upper() == request, handlers)

    if versions:
        handlers = [
            handler for handler in handlers
            if any(version in handler.versions for version in versions)
        ]

    return handlers


def sort_handlers(handlers, ascending=True):
    return sorted(
        handlers, key=lambda h: getattr(h, "index", 100000),
        reverse=not ascending
    )


def handler_supports_service(handler, service=None):
    """ Convenience method to check whether or not a handler supports a service.
    """
    if isinstance(handler.service, string_types):
        return handler.service.upper() == service
    else:
        return service in handler.service



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from textwrap import dedent

from django.test import TestCase, TransactionTestCase
from django.contrib.gis.geos import Polygon, MultiPolygon
from django.utils.six import assertCountEqual, b

from eoxserver.core.util import multiparttools as mp
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.core.config import get_eoxserver_config
from eoxserver.services.subset import Subsets, Trim, Slice
from eoxserver.services.result import result_set_from_raw_data
from eoxserver.resources.coverages import models


class MultipartTest(TestCase):
    """ Test class for multipart parsing/splitting
    """

    example_multipart = b(dedent("""\
        MIME-Version: 1.0\r
        Content-Type: multipart/mixed; boundary=frontier\r
        \r
        This is a message with multiple parts in MIME format.\r
        --frontier
        Content-Type: text/plain\r
        \r
        This is the body of the message.\r
        --frontier
        Content-Type: application/octet-stream\r
        Content-Transfer-Encoding: base64\r
        \r
        PGh0bWw+CiAgPGhlYWQ+CiAgPC9oZWFkPgogIDxib2R5PgogICAgPHA+VGhpcyBpcyB0aGUgYm9keSBvZiB0aGUgbWVzc2FnZS48L3A+CiAgPC9ib2R5Pgo8L2h0bWw+Cg==\r
        --frontier--
    """))

    def test_multipart_iteration(self):
        parsed = [
            (i[0], i[1].tobytes() if isinstance(i[1], memoryview) else i[1])
            for i in mp.iterate(self.example_multipart)
        ]

        self.assertEqual([
                ({b"MIME-Version": b"1.0", b"Content-Type": b"multipart/mixed; boundary=frontier"}, b""),
                ({b"Content-Type": b"text/plain"}, b"This is the body of the message."),
                ({b"Content-Type": b"application/octet-stream", b"Content-Transfer-Encoding": b"base64"}, b"PGh0bWw+CiAgPGhlYWQ+CiAgPC9oZWFkPgogIDxib2R5PgogICAgPHA+VGhpcyBpcyB0aGUgYm9keSBvZiB0aGUgbWVzc2FnZS48L3A+CiAgPC9ib2R5Pgo8L2h0bWw+Cg==")
            ], parsed
        )


class ResultSetTestCase(TestCase):

    example_multipart = b(dedent("""\
        MIME-Version: 1.0\r
        Content-Type: multipart/mixed; boundary=frontier\r
        \r
        This is a message with multiple parts in MIME format.\r
        --frontier
        Content-Type: text/plain\r
        Content-Disposition: attachmet; filename="message.msg"\r
        Content-Id: message-part\r
        \r
        This is the body of the message.\r
        --frontier
        Content-Type: application/octet-stream\r
        Content-Transfer-Encoding: base64\r
        \r
        PGh0bWw+CiAgPGhlYWQ+CiAgPC9oZWFkPgogIDxib2R5PgogICAgPHA+VGhpcyBpcyB0aGUgYm9keSBvZiB0aGUgbWVzc2FnZS48L3A+CiAgPC9ib2R5Pgo8L2h0bWw+Cg==\r
        --frontier--
    """))

    def test_result_set_from_raw(self):
        result_set = result_set_from_raw_data(self.example_multipart)
        self.assertTrue(len(result_set) == 2)

        first = result_set[0]
        second = result_set[1]
        
        if isinstance(first.data, str):
            first_data = b(first.data)
        else:
            first_data = first.data

        if isinstance(second.data, str):
            second_data = b(second.data)
        else:
            second_data = second.data
        
        self.assertEqual(
            first_data,
            b"This is the body of the message."
        )
        self.assertEqual(
            first.content_type,
            b"text/plain"
        )
        self.assertEqual(
            first.filename,
            b"message.msg"
        )
        self.assertEqual(
            first.identifier,
            b"message-part"
        )
        self.assertEqual(
            second_data,
            b"PGh0bWw+CiAgPGhlYWQ+CiAgPC9oZWFkPgogIDxib2R5PgogICAgPHA+VGhpcyBpcyB0aGUgYm9keSBvZiB0aGUgbWVzc2FnZS48L3A+CiAgPC9ib2R5Pgo8L2h0bWw+Cg=="
        )


class TemporalSubsetsTestCase(TransactionTestCase):
    def setUp(self):
        """ Set up a couple of test datasets to be distributed along the time
        axis as such:
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
        """
        p = parse_iso8601

        grid = models.Grid.objects.create(
            coordinate_reference_system='EPSG:4326',
            axis_1_name='long',
            axis_2_name='lat',
            axis_1_type=0,
            axis_2_type=0,
            axis_1_offset=5/100,
            axis_2_offset=5/100,
        )

        coverage_type = models.CoverageType.objects.create(name="RGB")
        self.datasets = [
            models.Coverage.objects.create(
                identifier="A",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:00Z"),
                end_time=p("2000-01-01T00:00:05Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="B",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:10Z"),
                end_time=p("2000-01-01T00:00:15Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="C",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:12Z"),
                end_time=p("2000-01-01T00:00:17Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="D",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:15Z"),
                end_time=p("2000-01-01T00:00:20Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="E",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:18Z"),
                end_time=p("2000-01-01T00:00:23Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="F",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:20Z"),
                end_time=p("2000-01-01T00:00:25Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="G",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:30Z"),
                end_time=p("2000-01-01T00:00:35Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            ),
            models.Coverage.objects.create(
                identifier="H",
                footprint=MultiPolygon(Polygon.from_bbox((0, 0, 5, 5))),
                begin_time=p("2000-01-01T00:00:40Z"),
                end_time=p("2000-01-01T00:00:40Z"),
                grid=grid,
                axis_1_size=100, axis_2_size=100,
                axis_1_origin=0, axis_2_origin=0,
                coverage_type=coverage_type
            )
        ]

        self.config = get_eoxserver_config()
        self.original_interval_interpretation = self.config.get(
            "services.owscommon", "time_interval_interpretation"
        )

        self.set_interpretation("closed")

    def tearDown(self):
        self.set_interpretation()

    def set_interpretation(self, value=None):
        self.config.set(
            "services.owscommon", "time_interval_interpretation",
            value if value else self.original_interval_interpretation
        )

    def evaluate_subsets(self, subsets, containment, expected_ids):
        """ Evaluates the subset via QuerySet filter and via the matches()
            functionality.
        """
        qs = models.EOObject.objects.filter(**subsets.get_filters(containment))

        # test the "filter()" function
        self.assertCountEqual(
            expected_ids, qs.values_list("identifier", flat=True)
        )

        # test the "matches()" function
        self.assertCountEqual(
            expected_ids, [
                ds.identifier for ds in self.datasets
                if subsets.matches(ds, containment)
            ]
        )

    def make_subsets(self, begin, end=None):
        if end is None:
            return Subsets([Slice("t", parse_iso8601(begin))])
        else:
            return Subsets([Trim(
                "t", parse_iso8601(begin), parse_iso8601(end)
            )])

    def test_all_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
        ]-------------------------------[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:00Z", "2000-01-01T00:00:35Z"
            ),
            "overlaps", ("A", "B", "C", "D", "E", "F", "G")
        )

    def test_all_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
        [-------------------------------]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:00Z", "2000-01-01T00:00:35Z"
            ),
            "overlaps", ("A", "B", "C", "D", "E", "F", "G")
        )

    def test_all_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
        ]...............................[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:00Z", "2000-01-01T00:00:35Z"
            ),
            "contains", ("B", "C", "D", "E", "F")
        )

    def test_all_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
        [...............................]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:00Z", "2000-01-01T00:00:35Z"
            ),
            "contains", ("A", "B", "C", "D", "E", "F", "G")
        )

    def test_middle_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
            ]-----------------------[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:05Z", "2000-01-01T00:00:30Z"
            ),
            "overlaps", ("B", "C", "D", "E", "F")
        )

    def test_middle_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
            [-----------------------]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:05Z", "2000-01-01T00:00:30Z"
            ),
            "overlaps", ("A", "B", "C", "D", "E", "F", "G")
        )

    def test_middle_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
            ].......................[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:05Z", "2000-01-01T00:00:30Z"
            ),
            "contains", ("B", "C", "D", "E", "F")
        )

    def test_middle_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
            [.......................]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:05Z", "2000-01-01T00:00:30Z"
            ),
            "contains", ("B", "C", "D", "E", "F")
        )

    def test_small_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                      ]---[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:15Z", "2000-01-01T00:00:20Z"
            ),
            "overlaps", ("C", "D", "E")
        )

    def test_small_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                      [---]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:15Z", "2000-01-01T00:00:20Z"
            ),
            "overlaps", ("B", "C", "D", "E", "F")
        )

    def test_small_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                      ]...[
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:15Z", "2000-01-01T00:00:20Z"
            ),
            "contains", ()
        )

    def test_small_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                      [...]
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:15Z", "2000-01-01T00:00:20Z"
            ),
            "contains", ("D",)
        )

    def test_point_1_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:30Z", "2000-01-01T00:00:30Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_1_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:30Z", "2000-01-01T00:00:30Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_1_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:30Z", "2000-01-01T00:00:30Z"
            ),
            "contains", ()
        )

    def test_point_1_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:30Z", "2000-01-01T00:00:30Z"
            ),
            "contains", ()
        )

    def test_point_2_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:32Z", "2000-01-01T00:00:32Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_2_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:32Z", "2000-01-01T00:00:32Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_2_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:32Z", "2000-01-01T00:00:32Z"
            ),
            "contains", ()
        )

    def test_point_2_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:35Z", "2000-01-01T00:00:35Z"
            ),
            "contains", ()
        )

    def test_point_3_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:35Z", "2000-01-01T00:00:35Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_3_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:35Z", "2000-01-01T00:00:35Z"
            ),
            "overlaps", ("G",)
        )

    def test_point_3_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:35Z", "2000-01-01T00:00:35Z"
            ),
            "contains", ()
        )

    def test_point_3_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:35Z", "2000-01-01T00:00:35Z"
            ),
            "contains", ()
        )

    def test_point_4_overlaps_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:40Z", "2000-01-01T00:00:40Z"
            ),
            "overlaps", ("H",)
        )

    def test_point_4_overlaps_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:40Z", "2000-01-01T00:00:40Z"
            ),
            "overlaps", ("H",)
        )

    def test_point_4_contains_open(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:40Z", "2000-01-01T00:00:40Z"
            ),
            "contains", ("H",)
        )

    def test_point_4_contains_closed(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.evaluate_subsets(
            self.make_subsets(
                "2000-01-01T00:00:40Z", "2000-01-01T00:00:40Z"
            ),
            "contains", ("H",)
        )

    def test_slice_1_overlaps(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.set_interpretation("open")
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:30Z"), "overlaps", ("G",)
        )

    def test_slice_1_contains(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                    |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:30Z"), "contains", ("G",)
        )

    def test_slice_2_overlaps(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:32Z"), "overlaps", ("G",)
        )

    def test_slice_2_contains(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                      |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:32Z"), "contains", ("G",)
        )

    def test_slice_3_overlaps(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:35Z"), "overlaps", ("G",)
        )

    def test_slice_3_contains(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                        |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:35Z"), "contains", ("G",)
        )

    def test_slice_4_overlaps(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:40Z"), "overlaps", ("H",)
        )

    def test_slice_4_contains(self):
        """
        |-A-|     |-B-|-D-|-F-|     |-G-|     H
                    |-C-||-E-|
                                              |
        """
        self.evaluate_subsets(
            self.make_subsets("2000-01-01T00:00:40Z"), "contains", ("H",)
        )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from django.conf import settings

from eoxserver.core import Component, implements
from eoxserver.services.result import ResultBuffer
from eoxserver.services.ows.version import Version
from eoxserver.services.ows.wcs.v20.encoders import WCS20EOXMLEncoder
from eoxserver.services.ows.wcs.interfaces import (
    WCSCoverageDescriptionRendererInterface
)


class NativeWCS20CoverageDescriptionRenderer(Component):
    """ Coverage description renderer for WCS 2.0 using the EO application
        profile.
    """

    implements(WCSCoverageDescriptionRendererInterface)

    versions = (Version(2, 0),)

    def supports(self, params):
        return params.version in self.versions

    def render(self, params):
        encoder = WCS20EOXMLEncoder()
        return [
            ResultBuffer(
                encoder.serialize(
                    encoder.encode_coverage_descriptions(params.coverages),
                    pretty_print=settings.DEBUG
                ),
                encoder.content_type
            )
        ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.services.ows.wcs.parameters import (
    CoverageRenderParams, CoverageDescriptionRenderParams
)
from eoxserver.services.subset import Subsets, Trim


class WCS11CoverageDescrptionRenderParams(CoverageDescriptionRenderParams):
    coverage_ids_key_name = "identifier"

    def __init__(self, coverages):
        super(WCS11CoverageDescrptionRenderParams, self).__init__(
            coverages, "1.1.2"
        )


class WCS11CoverageRenderParams(CoverageRenderParams):
    def __init__(self, coverage, bbox, format, gridcs=None, gridbasecrs=None, 
                 gridtype=None, gridorigin=None, gridoffsets=None):

        super(WCS11CoverageRenderParams, self).__init__(coverage, "1.1.2")
        self._bbox = list(bbox)
        self._format = format
        self._gridcs = gridcs
        self._gridbasecrs = gridbasecrs
        self._gridtype = gridtype
        self._gridorigin = gridorigin
        self._gridoffsets = gridoffsets

    coverage_id_key_name = "identifier"

    bbox        = property(lambda self: self._bbox)
    format      = property(lambda self: self._format)
    gridcs      = property(lambda self: self._gridcs)
    gridbasecrs = property(lambda self: self._gridbasecrs)
    gridtype    = property(lambda self: self._gridtype)
    gridorigin  = property(lambda self: self._gridorigin)
    gridoffsets = property(lambda self: self._gridoffsets)

    @property
    def subsets(self):
        return Subsets((
            Trim("x", self._bbox[0], self._bbox[2]),
            Trim("y", self._bbox[1], self._bbox[3]),
        ), crs=self._bbox[4])


    def __iter__(self):
        for k, v in super(WCS11CoverageRenderParams, self).__iter__():
            yield k, v

        yield ("boundingbox", ",".join(map(str, self.bbox)))
        yield ("format", self.format)
        
        if self.gridcs:
            yield ("gridcs", self.gridcs)
        
        if self.gridbasecrs:
            yield ("gridbasecrs", self.gridbasecrs)
        
        if self.gridoffsets:
            yield ("gridoffsets", ",".join(map(str, self.gridoffsets)))
        
        if self.gridtype is not None:
            yield ("gridtype", self.gridtype)
        
        if self.gridorigin is not None:
            yield ("gridorigin", ",".join(map(str, self.gridorigin)))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-04 14:26
from __future__ import unicode_literals

from django.db import migrations, models
import eoxserver.resources.coverages.models


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0004_auto_20181220_1300'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eoobject',
            name='identifier',
            field=models.CharField(max_length=256, unique=True, validators=[eoxserver.resources.coverages.models.eo_object_identifier_validator]),
        ),
    ]



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.decoders import kvp, xml, typelist
from eoxserver.services.ows.wps.util import get_processes
from eoxserver.services.ows.wps.exceptions import NoSuchProcessError
from eoxserver.services.ows.wps.v10.encoders import (
    WPS10ProcessDescriptionsXMLEncoder
)
from eoxserver.services.ows.wps.v10.util import nsmap


class WPS10DescribeProcessHandler(object):
    """ WPS 1.0 DescribeProcess service handler. """

    service = "WPS"
    versions = ("1.0.0",)
    request = "DescribeProcess"
    methods = ['GET', 'POST']

    @staticmethod
    def get_decoder(request):
        """ Get the WPS request decoder. """
        if request.method == "GET":
            return WPS10DescribeProcessKVPDecoder(request.GET)
        else:
            return WPS10DescribeProcessXMLDecoder(request.body)


    def handle(self, request):
        """ Handle HTTP request. """
        decoder = self.get_decoder(request)
        identifiers = set(decoder.identifiers)

        used_processes = []
        for process in get_processes():
            process_identifier = (
                getattr(process, 'identifier', None) or type(process).__name__
            )
            if process_identifier in identifiers:
                identifiers.remove(process_identifier)
                used_processes.append(process)

        for identifier in identifiers:
            raise NoSuchProcessError(identifier)

        encoder = WPS10ProcessDescriptionsXMLEncoder()
        return encoder.serialize(
            encoder.encode_process_descriptions(used_processes)
        )


class WPS10DescribeProcessKVPDecoder(kvp.Decoder):
    """ WPS 1.0 DescribeProcess HTTP/GET KVP request decoder. """
    #pylint: disable=too-few-public-methods
    identifiers = kvp.Parameter("identifier", type=typelist(str, ","))


class WPS10DescribeProcessXMLDecoder(xml.Decoder):
    """ WPS 1.0 DescribeProcess HTTP/POST XML request decoder. """
    #pylint: disable=too-few-public-methods
    identifiers = xml.Parameter("ows:Identifier/text()", num="+")
    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""\
This module contains a set of handler base classes which shall help to implement
a specific handler. Interface methods need to be overridden in order to work,
default methods can be overidden.
"""
import math
import re

from django.conf import settings
from django.db.models import Q
from django.urls import reverse
from django.http import HttpResponse

from eoxserver.core.decoders import kvp, typelist, InvalidParameterException
from eoxserver.core.config import get_eoxserver_config
from eoxserver.render.map.renderer import (
    get_map_renderer, # get_feature_info_renderer
)
from eoxserver.render.map.objects import Map
from eoxserver.resources.coverages import crss
from eoxserver.resources.coverages import models
from eoxserver.services.ows.wms.util import parse_bbox, parse_time, int_or_str
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.wms.exceptions import InvalidCRS
from eoxserver.services.ecql import parse, to_filter, get_field_mapping_for_model
from eoxserver.services import filters
from eoxserver.services.ows.wms.layermapper import LayerMapper
from eoxserver.services import views


class WMSBaseGetCapabilitiesHandler(object):
    """ Base for WMS capabilities handlers.
    """

    service = "WMS"
    request = "GetCapabilities"

    methods = ["GET"]

    def get_decoder(self, request):
        return WMSBaseGetCapbilitiesDecoder(request.GET)

    def handle(self, request):
        decoder = self.get_decoder(request)
        qs = models.EOObject.objects.all()

        cql_text = decoder.cql
        if cql_text:
            mapping, mapping_choices = filters.get_field_mapping_for_model(qs.model)
            ast = parse(cql_text)
            filter_expressions = to_filter(ast, mapping, mapping_choices)
            qs = qs.filter(filter_expressions)

        else:
            # lookup Collections, Products and Coverages
            qs = models.EOObject.objects.filter(
                Q(  # include "WMS-visible" Products
                    product__isnull=False,
                    service_visibility__service='wms',
                    service_visibility__visibility=True
                ) | Q(  # include "WMS-visible" Coverages
                    coverage__isnull=False,
                    service_visibility__service='wms',
                    service_visibility__visibility=True
                ) | Q(  # include all Collections, exclude "WMS-invisible" later
                    collection__isnull=False
                )
            ).exclude(
                collection__isnull=False,
                service_visibility__service='wms',
                service_visibility__visibility=False
            )

        qs = qs.select_subclasses()

        #
        map_renderer = get_map_renderer()
        raster_styles = map_renderer.get_raster_styles()
        geometry_styles = map_renderer.get_geometry_styles()

        layer_mapper = LayerMapper(
            map_renderer.get_supported_layer_types(), "__"
        )
        layer_descriptions = [
            layer_mapper.get_layer_description(
                eo_object, raster_styles, geometry_styles
            )
            for eo_object in qs
        ]

        encoder = self.get_encoder()
        conf = CapabilitiesConfigReader(get_eoxserver_config())
        return encoder.serialize(
            encoder.encode_capabilities(
                conf, request.build_absolute_uri(reverse(views.ows)),
                crss.getSupportedCRS_WMS(format_function=crss.asShortCode),
                map_renderer.get_supported_formats(), [],
                layer_descriptions
            ),
            pretty_print=settings.DEBUG
        ), encoder.content_type


class WMSBaseGetMapHandler(object):
    methods = ['GET']
    service = "WMS"
    request = "GetMap"

    def handle(self, request):
        decoder = self.get_decoder(request)

        minx, miny, maxx, maxy = decoder.bbox
        time = decoder.time
        crs = decoder.srs
        layer_names = decoder.layers

        width = decoder.width
        height = decoder.height

        # calculate the zoomlevel
        zoom = calculate_zoom((minx, miny, maxx, maxy), width, height, crs)

        if not layer_names:
            raise InvalidParameterException("No layers specified", "layers")

        srid = crss.parseEPSGCode(
            crs, (crss.fromShortCode, crss.fromURN, crss.fromURL)
        )
        if srid is None:
            raise InvalidCRS(crs, "crs")

        field_mapping, mapping_choices = get_field_mapping_for_model(
            models.Product
        )

        filter_expressions = filters.bbox(
            filters.attribute('footprint', field_mapping),
            minx, miny, maxx, maxy, crs, bboverlaps=False
        )

        if time:
            filter_expressions &= filters.time_interval(time)

        cql = getattr(decoder, 'cql', None)
        if cql:
            cql_filters = to_filter(
                parse(cql), field_mapping, mapping_choices
            )
            filter_expressions &= cql_filters

        # TODO: multiple sorts per layer?
        sort_by = getattr(decoder, 'sort_by', None)
        if sort_by:
            sort_by = (field_mapping.get(sort_by[0], sort_by[0]), sort_by[1])

        styles = decoder.styles

        if styles:
            styles = styles.split(',')
        else:
            styles = [None] * len(layer_names)

        dimensions = {
            "time": time,
            "elevation": decoder.elevation,
            "ranges": decoder.dim_range,
            "bands": decoder.dim_bands,
            "wavelengths": decoder.dim_wavelengths,
        }

        map_renderer = get_map_renderer()

        layer_mapper = LayerMapper(
            map_renderer.get_supported_layer_types(), "__"
        )

        layers = []
        for layer_name, style in zip(layer_names, styles):
            name, suffix = layer_mapper.split_layer_suffix_name(layer_name)
            layer = layer_mapper.lookup_layer(
                name, suffix, style,
                filter_expressions, sort_by, zoom=zoom, **dimensions
            )
            layers.append(layer)

        map_ = Map(
            width=decoder.width, height=decoder.height, format=decoder.format,
            bbox=(minx, miny, maxx, maxy), crs=crs,
            bgcolor=decoder.bgcolor, transparent=decoder.transparent,
            layers=layers
        )

        result_bytes, content_type, filename = map_renderer.render_map(map_)

        response = HttpResponse(result_bytes, content_type=content_type)
        if filename:
            response['Content-Disposition'] = 'inline; filename="%s"' % filename

        return response


class WMSBaseGetFeatureInfoHandler(object):
    methods = ['GET']
    service = "WMS"
    request = "GetFeatureInfo"

    def handle(self, request):
        decoder = self.get_decoder(request)

        minx, miny, maxx, maxy = decoder.bbox
        x = decoder.x
        y = decoder.y
        time = decoder.time
        crs = decoder.srs
        layer_names = decoder.layers

        width = decoder.width
        height = decoder.height

        # calculate the zoomlevel
        zoom = calculate_zoom((minx, miny, maxx, maxy), width, height, crs)

        if not layer_names:
            raise InvalidParameterException("No layers specified", "layers")

        srid = crss.parseEPSGCode(
            crs, (crss.fromShortCode, crss.fromURN, crss.fromURL)
        )
        if srid is None:
            raise InvalidCRS(crs, "crs")

        field_mapping, mapping_choices = get_field_mapping_for_model(
            models.Product
        )

        # calculate resolution
        # TODO: dateline
        resx = (maxx - minx) / width
        resy = (maxy - miny) / height

        p_minx = x * resx
        p_miny = y * resy
        p_maxx = (x + 1) * resx
        p_maxy = (y + 1) * resy

        filter_expressions = filters.bbox(
            filters.attribute('footprint', field_mapping),
            p_minx, p_miny, p_maxx, p_maxy, crs, bboverlaps=False
        )

        if time:
            filter_expressions &= filters.time_interval(time)

        cql = getattr(decoder, 'cql', None)
        if cql:
            cql_filters = to_filter(
                parse(cql), field_mapping, mapping_choices
            )
            filter_expressions &= cql_filters

        # TODO: multiple sorts per layer?
        sort_by = getattr(decoder, 'sort_by', None)
        if sort_by:
            sort_by = (field_mapping.get(sort_by[0], sort_by[0]), sort_by[1])

        styles = decoder.styles

        if styles:
            styles = styles.split(',')
        else:
            styles = [None] * len(layer_names)

        dimensions = {
            "time": time,
            "elevation": decoder.elevation,
            "ranges": decoder.dim_range,
            "bands": decoder.dim_bands,
            "wavelengths": decoder.dim_wavelengths,
        }

        feature_info_renderer = get_feature_info_renderer()

        layer_mapper = LayerMapper(
            feature_info_renderer.get_supported_layer_types(), "__"
        )

        layers = []
        for layer_name, style in zip(layer_names, styles):
            name, suffix = layer_mapper.split_layer_suffix_name(layer_name)
            layer = layer_mapper.lookup_layer(
                name, suffix, style,
                filter_expressions, sort_by, zoom=zoom, **dimensions
            )
            layers.append(layer)

        map_ = Map(
            width=decoder.width, height=decoder.height, format=decoder.format,
            bbox=(minx, miny, maxx, maxy), crs=crs,
            bgcolor=decoder.bgcolor, transparent=decoder.transparent,
            layers=layers
        )

        result_bytes, content_type, filename = feature_info_renderer.render_feature_info(map_)

        response = HttpResponse(result_bytes, content_type=content_type)
        if filename:
            response['Content-Disposition'] = 'inline; filename="%s"' % filename

        return response


class WMSBaseGetCapbilitiesDecoder(kvp.Decoder):
    cql = kvp.Parameter(num="?")


def parse_transparent(value):
    value = value.upper()
    if value == 'TRUE':
        return True
    elif value == 'FALSE':
        return False
    raise ValueError("Invalid value for 'transparent' parameter.")


def parse_ranges(value):
    ranges_separator = getattr(settings, 'EOXS_WMS_DIM_RANGES_SEPARATOR', r',')
    range_separator = getattr(settings, 'EOXS_WMS_DIM_RANGE_SEPARATOR', r'\s+')
    return [
        [
            float(v)
            for v in re.split(range_separator, rng.strip())
        ]
        for rng in re.split(ranges_separator, value)
    ]


def parse_sort_by(value):
    items = value.strip().split()
    assert items[1] in ['A', 'D']
    return (items[0], 'ASC' if items[1] == 'A' else 'DESC')


class WMSBaseGetMapDecoder(kvp.Decoder):
    layers = kvp.Parameter(type=typelist(str, ","), num=1)
    styles = kvp.Parameter(num="?")
    width = kvp.Parameter(num=1)
    height = kvp.Parameter(num=1)
    format = kvp.Parameter(num=1)
    bgcolor = kvp.Parameter(num='?')
    transparent = kvp.Parameter(num='?', default=False, type=parse_transparent)

    bbox = kvp.Parameter('bbox', type=parse_bbox, num=1)
    srs = kvp.Parameter(num=1)

    time = kvp.Parameter(type=parse_time, num="?")
    elevation = kvp.Parameter(type=float, num="?")
    dim_bands = kvp.Parameter(type=typelist(int_or_str, ","), num="?")
    dim_wavelengths = kvp.Parameter(type=typelist(float, ","), num="?")
    dim_range = kvp.Parameter(type=parse_ranges, num="?")

    cql = kvp.Parameter(num="?")

    sort_by = kvp.Parameter('sortBy', type=parse_sort_by, num="?")


def calculate_zoom(bbox, width, height, crs):
    # TODO: make this work for other CRSs
    lon_diff = bbox[2] - bbox[0]
    lat_diff = bbox[3] - bbox[1]

    max_diff = max(lon_diff, lat_diff)
    if max_diff < (360 / pow(2, 20)):
        return 21
    else:
        zoom = int(-1 * (math.log(max_diff, 2) - (math.log(360, 2))))
        if zoom < 1:
            zoom = 1
        return zoom
